# -*- coding: utf-8 -*-
from datetime import datetime, timedelta
from typing import Optional
from fastapi import Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer
from jose import JWTError, jwt
from passlib.context import CryptContext

# Custom imports
from innovationmerge.app.schemas.user_schema import UserInDB, TokenData, User
from innovationmerge.config import configs

# to get a string like this run:
# openssl rand -hex 32
SECRET_KEY = configs.SECRET_KEY
ALGORITHM = configs.ALGORITHM
ACCESS_TOKEN_EXPIRE_MINUTES = configs.ACCESS_TOKEN_EXPIRE_MINUTES

fake_users_db = {
    "user1": {
        "username": "user1",
        "full_name": "user1",
        "email": "user1@example.com",
        "hashed_password": (
            "$2b$12$foF6N3eZTMP2m9Ugoy2b8uof6uoaEAroTho2wgrLkUyY3oWiiTBEK"
        ),
        "disabled": False,
    },
    "user2": {
        "username": "user2",
        "full_name": "user2",
        "email": "user2@example.com",
        "hashed_password": (
            "$2b$12$fgGOw0lDMLaFBlO3K93Gku7paGO89aAle2BO.pKc62PbdZ2/g83WC"
        ),
        "disabled": False,
    },
}

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/authenticate")


def verify_password(plain_password, hashed_password):
    return pwd_context.verify(plain_password, hashed_password)


def get_password_hash(password):
    return pwd_context.hash(password)


def get_user(db, username: str):
    if username in db:
        user_dict = db[username]
        return UserInDB(**user_dict)


def authenticate_user(fake_db, username: str, password: str):
    user = get_user(fake_db, username)
    if not user:
        return False
    if not verify_password(password, user.hashed_password):
        return False
    return user


def create_access_token(data: dict, expires_delta: Optional[timedelta] = None):
    to_encode = data.copy()
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(minutes=15)
    to_encode.update({"exp": expire})
    encoded_jwt = jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)
    return encoded_jwt


async def get_current_user(token: str = Depends(oauth2_scheme)):
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    try:
        payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        username: str = payload.get("sub")
        if username is None:
            raise credentials_exception
        token_data = TokenData(username=username)
    except JWTError:
        raise credentials_exception
    user = get_user(fake_users_db, username=token_data.username)
    if user is None:
        raise credentials_exception
    return user


async def get_current_active_user(
    current_user: User = Depends(get_current_user),
):
    if current_user.disabled:
        raise HTTPException(status_code=400, detail="Inactive user")
    return current_user



from fastapi import Depends, HTTPException, status, APIRouter
from fastapi.security import OAuth2PasswordRequestForm
from datetime import timedelta

# Custom imports
from innovationmerge.app.schemas.user_schema import Token, User
from innovationmerge.app.security import (
    authenticate_user,
    fake_users_db,
    create_access_token,
    get_current_active_user,
)
from innovationmerge.config import configs

router = APIRouter()

ACCESS_TOKEN_EXPIRE_MINUTES = configs.ACCESS_TOKEN_EXPIRE_MINUTES


@router.post("/authenticate", response_model=Token)
async def login_for_access_token(
    form_data: OAuth2PasswordRequestForm = Depends(),
):
    user = authenticate_user(
        fake_users_db, form_data.username, form_data.password
    )
    if not user:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )
    access_token_expires = timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = create_access_token(
        data={"sub": user.username}, expires_delta=access_token_expires
    )
    return {"access_token": access_token, "token_type": "bearer"}


@router.get("/users/me/", response_model=User)
async def read_users_me(current_user: User = Depends(get_current_active_user)):
    return current_user



from fastapi import HTTPException, APIRouter, status, Response, Depends
from innovationmerge.app.schemas.sample_schema import SampleItem
from innovationmerge.docs.api_doc import sample_api_doc
from innovationmerge.app.sample import sample_division
from innovationmerge.app.schemas.user_schema import User
from innovationmerge.app.security import get_current_active_user

router = APIRouter()


@router.post("/sample_api", responses=sample_api_doc)
async def sample_api_function(
    response: Response,
    sample_item: SampleItem,
    current_user: User = Depends(get_current_active_user)
):
    try:
        result = sample_division(sample_item.val1, sample_item.val2)
        status_data = {
            "status": "Success",
            "status_code": status.HTTP_200_OK,
            "message": "No Exception",
            "data": result,
        }
        response.status_code = status.HTTP_200_OK
    except Exception as e:
        if type(e).__name__ == "ZeroDivisionError":
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Cannot divide by Zero"
            )
        else:
            status_data = {
                "status": "Failed",
                "status_code": status.HTTP_404_NOT_FOUND,
                "message": "Data Not Available",
                "data": [],
            }
            print(type(e).__name__)
            response.status_code = status.HTTP_404_NOT_FOUND
    return status_data



test_api_doc = {
    200: {
        "description": "Successful Response",
        "content": {"application/json": {"example": {"Status": "Working"}}},
    }
}

sample_api_doc = {
    200: {
        "description": "Successful Response",
        "content": {
            "application/json": {
                "example": {
                    "status": "Success",
                    "status_code": 200,
                    "message": "No Exception",
                    "data": 1,
                }
            }
        },
    },
    400: {
        "description": "Bad Request",
        "content": {
            "application/json": {
                "example": {"detail": "Cannot divide by Zero"}
            }
        },
    },
    404: {
        "description": "Not Found",
        "content": {
            "application/json": {
                "example": {
                    "status": "Failed",
                    "status_code": 404,
                    "message": "Data Not Available",
                    "data": [],
                }
            }
        },
    },
    401: {
        "description": "HTTP_401_Unauthorized",
        "content": {
            "application/json": {"example": {"detail": "Not authenticated"}}
        },
    },
}



# -*- coding: utf-8 -*-
import logging
import sys
import os
from pythonjsonlogger import jsonlogger

# Custom imports

cwd = os.path.dirname(os.path.realpath(__file__))
sys.path.append(cwd)

# Create a custom logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def log_init():
    # Create Handler
    screenHandler = logging.StreamHandler(stream=sys.stdout)

    # Add a suffix which you want
    screenHandler.suffix = "%Y%m%d"

    # Create formatters and add it to handlers
    formatter = jsonlogger.JsonFormatter(
        fmt="%(asctime)s | %(levelname)8s | %(message)s",
        datefmt="%Y/%m/%d %H:%M:%S"
    )
    screenHandler.setFormatter(formatter)

    # Add Handler
    logger.addHandler(screenHandler)




# Python Imports
from fastapi.testclient import TestClient

# Custom imports
from innovationmerge.app import create_app
from app.sample import sample_division

# Creat Testing Client
app = create_app()
client = TestClient(app)


# Make sure to start function name with test
def test_sample_division():
    assert sample_division(1, 1) == 1


def test_user_token():
    # To check access_token functionality
    response = client.post(
        "/authenticate",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={"username": "user1", "password": "user1@123"},
    )
    assert response.status_code == 200
    auth_token = response.json().get("access_token")
    assert len(auth_token) > 25

    # To check Incorrect username or password functionality
    response = client.post(
        "/authenticate",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={"username": "user2@domain.com", "password": "user1@123"},
    )
    assert response.status_code == 401
    assert response.json() == {"detail": "Incorrect username or password"}


def test_user_me():
    # To check Not authenticated functionality
    response = client.get("/users/me/")
    assert response.status_code == 401
    assert response.json() == {"detail": "Not authenticated"}

    # To get access_token
    response = client.post(
        "/authenticate",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={"username": "user1", "password": "user1@123"},
    )
    assert response.status_code == 200
    auth_token = "Bearer " + response.json().get("access_token")

    # To check user information functionality
    response = client.get(
        "/users/me/",
        headers={"Authorization": auth_token},
    )
    assert response.status_code == 200
    data = response.json()
    assert data["email"] == "user1@example.com"


def test_test_api():
    response = client.get("/")
    assert response.status_code == 200
    assert response.json() == {"Status": "Working"}


def test_sample_api():
    response = client.post(
        "/authenticate",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={"username": "user1", "password": "user1@123"},
    )
    assert response.status_code == 200
    auth_token = "Bearer " + response.json().get("access_token")

    response = client.post(
        "/v1/sample_api",
        headers={
            "Content-Type": "application/json",
            "Authorization": auth_token,
        },
        json={"val1": 2, "val2": 2},
    )
    assert response.status_code == 200
    assert response.json() == {
        "status": "Success",
        "status_code": 200,
        "message": "No Exception",
        "data": 1,
    }

    response = client.post(
        "/v1/sample_api",
        headers={
            "Content-Type": "application/json",
            "Authorization": auth_token,
        },
        json={"val1": 0, "val2": 0},
    )
    assert response.status_code == 400
    assert response.json() == {"detail": "Cannot divide by Zero"}



import os
from fastapi.staticfiles import StaticFiles
from innovationmerge.app.routers import sample_route, authenticate
from innovationmerge.docs.api_doc import test_api_doc


def router_init(app):
    # Static Path
    static_data_dir = os.path.join("innovationmerge", "data")
    app.mount("/static", StaticFiles(directory=static_data_dir), name="static")


    # Testing route
    @app.get("/", responses=test_api_doc)
    def test_api():
        return {"Status": "Working"}

    # API routers
    app.include_router(authenticate.router, tags=["Security"])
    app.include_router(sample_route.router, tags=["Sample"], prefix="/v1")


from fastapi import FastAPI

# Custom imports
from innovationmerge.app.logger import log_init, logger
from innovationmerge.app.routers import router_init
from innovationmerge.config import configs
from innovationmerge.app.middleware import middleware_init


def conf_init(app):
    logger.info(msg=f"Starting app with {configs.ENVIRONMENT} environment")


async def start_event():
    logger.info("API Started")


async def shutdown_event():
    logger.info("API Stopped")


def create_app():
    # Create API 
    app = FastAPI(
        title="innovationmerge API's",
        description="innovationmerge core components",
        version="1.0",
        on_startup=[start_event],
        on_shutdown=[shutdown_event],
    )

    # Enable Logging
    log_init()

    # Load configuration
    conf_init(app)

    # Initialize the middleware
    middleware_init(app)

    # Initialize routing configuration
    router_init(app)

    return app



import cv2
import time


def read_image(input_image_path):
    cv2_image = cv2.imread(input_image_path)
    return cv2_image


def load_labels(filename):
  with open(filename, 'r') as f:
    return [line.strip() for line in f.readlines()]


def draw_detections(detections, cv2_image):
    for detection in detections.get("detection"):
       xmin = detection.get("label_bounding_box")[0].get("x")
       ymin = detection.get("label_bounding_box")[0].get("y")
       xmax = detection.get("label_bounding_box")[1].get("x")
       ymax = detection.get("label_bounding_box")[1].get("y")
       cv2.rectangle(cv2_image, (xmin, ymin), (xmax, ymax), (10, 255, 0), 2)
        # Draw label
       object_name = detection["label_class_name"]
       confidence = detection["confidence"]
       label = '%s: %d%%' % (object_name, int(confidence*100))
       labelSize, baseLine = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, 0.7, 2)
       label_ymin = max(ymin, labelSize[1] + 10)
       cv2.rectangle(cv2_image, (xmin, label_ymin-labelSize[1]-10), (xmin+labelSize[0], label_ymin+baseLine-10), (255, 255, 255), cv2.FILLED)
       cv2.putText(cv2_image, label, (xmin, label_ymin-7), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 0, 0), 2)
    return cv2_image


def save_detection_image(detections : dict, cv2_image, output_image_path: str, model_type: str = "ssd"):

    if model_type == "ssd":
        for detection in detections.get("detection"):
            xmin = detection.get("label_bounding_box")[0].get("x")
            ymin = detection.get("label_bounding_box")[0].get("y")
            xmax = detection.get("label_bounding_box")[1].get("x")
            ymax = detection.get("label_bounding_box")[1].get("y")
            cv2.rectangle(cv2_image, (xmin, ymin), (xmax, ymax), (10, 255, 0), 2)
            # Draw label
            object_name = detection["label_class_name"]
            confidence = detection["confidence"]
            label = '%s: %d%%' % (object_name, int(confidence*100))
            labelSize, baseLine = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, 0.7, 2)
            label_ymin = max(ymin, labelSize[1] + 10)
            cv2.rectangle(cv2_image, (xmin, label_ymin-labelSize[1]-10), (xmin+labelSize[0], label_ymin+baseLine-10), (255, 255, 255), cv2.FILLED)
            cv2.putText(cv2_image, label, (xmin, label_ymin-7), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 0, 0), 2)
        cv2.imwrite(output_image_path, cv2_image)
    else:
        position = (30, 30)
        font_scale = 0.75
        color = (255, 255, 255)
        thickness = 2
        font = cv2.FONT_HERSHEY_SIMPLEX
        line_type = cv2.LINE_AA
        x, y0 = position
        for i, line in enumerate(detections.get("detection")):
            text_size, _ = cv2.getTextSize(line, font, font_scale, thickness)
            line_height = text_size[1] + 5
            y = y0 + i * line_height
            cv2.putText(cv2_image,
                        line,
                        (x, y),
                        font,
                        font_scale,
                        color,
                        thickness,
                        line_type)
        cv2.imwrite(output_image_path, cv2_image)


def show_detection_image(cv2_image):
    cv2.imshow('Detection', cv2_image)
    # Press any key to continue to next image, or press 'q' to quit
    if cv2.waitKey(0) == ord('q'):
        time.sleep(10)
    # Clean up
    cv2.destroyAllWindows()




from fastapi.middleware.cors import CORSMiddleware
from fastapi import Response
import uuid
import time

# Custom imports
from innovationmerge.app.logger import logger


def middleware_init(app):
    # Cors
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    @app.middleware("http")
    async def log_request(request, call_next):
        start_time = time.time()
        request_id = str(uuid.uuid4().hex)
        logger.info(
            f"{request_id} {request.client.host} \
            {request.method} {request.url}"
        )
        response = await call_next(request)
        if response.status_code != 200:
            logger.error(f"{request_id} {response.status_code}")
        else:
            logger.info(f"{request_id} {response.status_code}")
        body = b""
        async for chunk in response.body_iterator:
            body += chunk
        process_time = (time.time() - start_time) * 1000
        response.headers["X-Process-Time-ms"] = str(process_time)
        return Response(
            content=body,
            status_code=response.status_code,
            headers=dict(response.headers),
            media_type=response.media_type,
        )



from pydantic import BaseModel
from typing import Optional


class Message(BaseModel):
    message: str


class Token(BaseModel):
    access_token: str
    token_type: str


class TokenData(BaseModel):
    username: Optional[str] = None


class User(BaseModel):
    username: str
    email: Optional[str] = None
    full_name: Optional[str] = None
    disabled: Optional[bool] = None


class UserInDB(User):
    hashed_password: str



import tensorflow as tf
import cv2
import numpy as np
import time
from innovationmerge.configurations.constants import DETECTION_RESPONSE_BB, DETECTION_RESPONSE

class EdgeObjectDetectionTfLite():
    """
    Object detection inference using Tensorflow Lite models. 
    This class includes loading the model, preparing input data, invoking the interpreter, and extracting output results. 
    """
    def __init__(self, model_file_path):
        # Load the TensorFlow Lite model
        self.interpreter = tf.lite.Interpreter(model_path=model_file_path)
        self.interpreter.allocate_tensors()

        # Set input normalization parameters
        self.input_mean = 127.5
        self.input_std = 127.5

    def detect(self, cv2_image, labels_list, threshold, model_type="ssd"):
        """
        Parameters:
            cv2_image: OpenCV 2 Image
            labels_list: labels list
            threshold: minimum score
        Returns:
            Array of Detections(Label, Bounding Box, Confidence score)
        """
        # # Get input and output details from the interpreter
        input_details = self.interpreter.get_input_details()
        output_details = self.interpreter.get_output_details()

        # Convert the input image to RGB format
        normalize_height = input_details[0]['shape'][1]
        normalize_width = input_details[0]['shape'][2]

        # convert input image to gray scale
        image_rgb = cv2.cvtColor(cv2_image, cv2.COLOR_BGR2RGB)
        input_img_height, input_img_width, _ = image_rgb.shape

        # Resize the input image to match the model's input requirements
        image_resized = cv2.resize(image_rgb, (normalize_width, normalize_height))

        # Add an extra dimension to the input data
        input_data = np.expand_dims(image_resized, axis=0)

        # Check if the model expects floating-point input
        floating_model = (input_details[0]['dtype'] == np.float32)
        if input_details[0]['dtype'] == type(np.float32(1.0)):
            floating_model = True

        # Normalize the input data if floating_model is True
        if floating_model:
            input_data = (np.float32(input_data) - self.input_mean) / self.input_std
        
        # Set the input tensor for the interpreter
        self.interpreter.set_tensor(input_details[0]['index'], input_data)

        # Perform the model inference
        start_time = time.time()
        self.interpreter.invoke()
        stop_time = time.time()

        # Determine the indices for accessing the output tensor.The specific indices for accessing the output tensors depend on whether the model is a TensorFlow 1.x or TensorFlow 2.x model.
        out_name = output_details[0]['name']
        if ('StatefulPartitionedCall' in out_name): # This is a TF2 model
            boxes_idx, classes_idx, scores_idx = 1, 3, 0
        else: # This is a TF1 model
            boxes_idx, classes_idx, scores_idx = 0, 1, 2

        if model_type == "ssd":
            # Retrieve detection results
            boxes = self.interpreter.get_tensor(output_details[boxes_idx]['index'])[0] # Bounding box coordinates of detected objects
            classes = self.interpreter.get_tensor(output_details[classes_idx]['index'])[0] # Class index of detected objects
            scores = self.interpreter.get_tensor(output_details[scores_idx]['index'])[0] # Confidence of detected objects

            # Create custom response
            detection_response = {}
            detection_list = []
            for i in range(len(scores)):
                label_dict = DETECTION_RESPONSE_BB.copy()
                if ((scores[i] > threshold) and (scores[i] <= 1.0)):
                    ymin = int(max(1,(boxes[i][0] * input_img_height)))
                    xmin = int(max(1,(boxes[i][1] * input_img_width)))
                    ymax = int(min(input_img_height,(boxes[i][2] * input_img_height)))
                    xmax = int(min(input_img_width,(boxes[i][3] * input_img_width)))
                    label_dict["confidence"] = scores[i]
                    label_dict["label_bounding_box"] = [{"x": xmin, "y": ymin}, {"x": xmax, "y": ymax}]
                    label_dict["label_class_name"] = labels_list[int(classes[i])]
                    detection_list.append(label_dict)
            processing_time = stop_time - start_time
            detection_response= {"detection": detection_list, "processing_time": processing_time}
            return detection_response
        else:
            output_data = self.interpreter.get_tensor(output_details[0]['index'])
            results = np.squeeze(output_data)
            top_k = results.argsort()[-5:][::-1]
            detection_list = []
            for i in top_k:
                print(labels_list[i])
                detection_list.append(' '.join(labels_list[i].split()[1:]))
            processing_time = stop_time - start_time
            detection_response = {"detection": detection_list, "processing_time": processing_time}
            return detection_response



# -*- coding: utf-8 -*-
class innovationmergeException(Exception):
    """Base class for exceptions raised by innovationmerge  modules"""


class innovationmergeMethodException(Exception):
    """Base class for exceptions raised by innovationmerge  modules"""
    def __init__(self, message):
        super().__init__(str(message))


# -*- coding: utf-8 -*-
from functools import lru_cache
from pydantic import BaseSettings, Field
from typing import Optional
from io import StringIO
import io
import os
from contextlib import contextmanager
from dotenv.main import DotEnv

cwd = os.getcwd()
root_path = os.path.join(cwd, "innovationmerge")


def my_get_stream(self):
    if isinstance(self.dotenv_path, StringIO):
        yield self.dotenv_path
    elif os.path.isfile(self.dotenv_path):
        with io.open(self.dotenv_path, encoding="utf-8") as stream:
            yield stream
    else:
        if self.verbose:
            pass
        yield StringIO("")


DotEnv._get_stream = contextmanager(my_get_stream)


class Settings(BaseSettings):
    ENVIRONMENT = os.getenv("ENVIRONMENT")

    class Config:
        ENVIRONMENT = os.getenv("ENVIRONMENT")
        env_file = os.path.join(root_path, ".env." + ENVIRONMENT)
        case_sensitive = True
        print("env_file", env_file)


class GetConfig(Settings):
    ENVIRONMENT: str = Field(None, env="ENVIRONMENT")
    HOST: str = Field(None, env="HOST")
    PORT: int = Field(None, env="PORT")
    SECRET_KEY: str = Field(None, env="SECRET_KEY")
    ALGORITHM: str = Field(None, env="ALGORITHM")
    ACCESS_TOKEN_EXPIRE_MINUTES: int = Field(None, env="ACCESS_TOKEN_EXPIRE_MINUTES")


class FactoryConfig:
    """Returns a config instance dependending on the ENV_STATE variable."""

    def __init__(self, env_state: Optional[str]):
        self.env_state = env_state

    def __call__(self):
        return GetConfig()


@lru_cache()
def get_configs():
    from dotenv import load_dotenv

    load_dotenv(encoding="utf-8", verbose=True)
    return FactoryConfig(Settings().ENVIRONMENT)()


configs = get_configs()
print("configs", configs.dict())



import os
from innovationmerge import cpuObjectDetectionTfLite, edgeObjectDetectionTfLite
from innovationmerge.src.utils.responses import load_labels, read_image

# Declare paths
cwd = os.getcwd()
model_path = os.path.join(cwd, "models", "ssd_mobilenet_v3_large_coco_2020_01_14")
model_file_path = os.path.join(model_path, "model.tflite")
labels_path = os.path.join(model_path, "labels.txt")
input_image_path = os.path.join(cwd, "data", "testdata", "images.jpg")


def test_cpu_object_detect():
    labels = load_labels(labels_path)
    cv2_image = read_image(input_image_path)
    detect_objects = cpuObjectDetectionTfLite(model_file_path)
    detection_result = detect_objects.detect(cv2_image, labels)
    print(detection_result)
    assert detection_result.get('detection')[0].get('label_class_name') == "cat"


def test_edge_object_detect():
    labels = load_labels(labels_path)
    cv2_image = read_image(input_image_path)
    detect_objects = edgeObjectDetectionTfLite(model_file_path)
    detection_result = detect_objects.detect(cv2_image, labels)
    print(detection_result)
    assert detection_result.get('detection')[0].get('label_class_name') == "cat"